import io
import logging
import os
import subprocess  # nosec: B404
import tarfile
from pathlib import Path
from typing import Any, Dict, Iterator, Optional, Tuple, cast

import testcontainers.core.container
from docker.errors import ImageNotFound
from docker.models.containers import Container
from docker.models.images import Image as DockerImage
from testcontainers.core.docker_client import DockerClient
from testcontainers.core.utils import inside_container


class DockerContainer(testcontainers.core.container.DockerContainer):
    def __init__(self, *args: Any, network: Optional[str] = None, **kwargs: Any) -> None:
        self.network = network or os.getenv("TESTCONTAINER_DOCKER_NETWORK") or "bridge"
        super().__init__(*args, **kwargs, network=self.network)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        logger = logging.getLogger(self.__class__.__name__)
        logs = self.get_wrapped_container().logs(timestamps=True).decode().split("\n")
        for log in logs:
            logger.info(log)
        self.stop()

    def get_container_host_ip(self) -> str:
        host = self.get_docker_client().host()
        if not host:
            return "localhost"

        if inside_container() and not os.getenv("DOCKER_HOST"):
            gateway_ip = self.get_container_gateway_ip()
            if gateway_ip == host:
                return self.get_container_internal_ip()
            return gateway_ip
        return host

    def get_container_internal_ip(self) -> str:
        container = self.get_docker_client().get_container(self.get_wrapped_container().id)
        return container["NetworkSettings"]["Networks"][self.network]["IPAddress"]

    def get_container_gateway_ip(self) -> str:
        container = self.get_docker_client().get_container(self.get_wrapped_container().id)
        return container["NetworkSettings"]["Networks"][self.network]["Gateway"]

    def restart_container(self) -> None:
        self.get_wrapped_container().restart()


class EphemeralDockerImage:
    """Builds a Docker image from a given Dockerfile and removes it when the context manager exits."""

    image: DockerImage

    def __init__(
        self,
        dockerfile: Optional[Path] = None,
        context: Optional[Path] = None,
        target: Optional[str] = None,
        docker_client_kw: Optional[Dict] = None,
    ) -> None:
        self.dockerfile = str(dockerfile) if dockerfile else None
        self.context = str(context) if context else "."
        self.target = target
        self.client = DockerClient(**(docker_client_kw or {}))

    def __enter__(self) -> DockerImage:
        self.image = self.build_image()
        return self.image

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.remove_image()

    def build_image(self) -> DockerImage:
        if os.getenv("DOCKER_BUILDKIT"):
            return self._build_with_docker_buildkit()
        return self._build_with_docker_client()

    def remove_image(self) -> None:
        self.client.client.images.remove(image=str(self.image.id))

    def _build_with_docker_buildkit(self) -> DockerImage:
        cmd = ["docker", "build", "-q", "--rm=true"]
        if self.dockerfile:
            cmd.extend(["-f", self.dockerfile])
        if self.target:
            cmd.extend(["--target", self.target])
        cmd.append(self.context)

        result = subprocess.run(
            cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True
        )  # nosec: B603
        image_id = result.stdout.decode("utf-8").strip()
        return cast(DockerImage, self.client.client.images.get(image_id))

    def _build_with_docker_client(self) -> DockerImage:
        image, _ = cast(
            Tuple[DockerImage, Iterator],
            self.client.client.images.build(
                dockerfile=self.dockerfile,
                path=self.context,
                target=self.target,
                forcerm=True,
            ),
        )
        return image


def get_docker_image(image_id: str, docker_client_kw: Optional[Dict] = None) -> DockerImage:
    client = DockerClient(**(docker_client_kw or {}))
    try:
        return cast(DockerImage, client.client.images.get(image_id))
    except ImageNotFound:
        return cast(DockerImage, client.client.images.pull(image_id))


def copy_folder_to_container(host_path: Path, container_path: Path, container: Container) -> None:
    tar_stream = io.BytesIO()
    with tarfile.open(fileobj=tar_stream, mode="w") as tar:
        for root, _, files in os.walk(host_path):
            for file in files:
                file_path = Path(root) / file
                arcname = os.path.relpath(file_path, host_path)
                tar.add(file_path, arcname=arcname)
    tar_stream.seek(0)
    container.put_archive(path=container_path, data=tar_stream)



import json
from typing import Any, Dict, List, Protocol, Type, TypeVar, Union

from types_aiobotocore_sns import SNSClient
from types_aiobotocore_sqs import SQSClient
from types_aiobotocore_sqs.literals import QueueAttributeFilterType, QueueAttributeNameType

MessageType = TypeVar("MessageType")


class TomodachiSNSSQSEnvelopeStatic(Protocol):
    @classmethod
    async def build_message(
        cls: "TomodachiSNSSQSEnvelopeStatic", service: Any, topic: str, data: Any, **kwargs: Any
    ) -> str:
        ...

    @classmethod
    async def parse_message(cls: "TomodachiSNSSQSEnvelopeStatic", payload: str, **kwargs: Any) -> Union[dict, tuple]:
        ...


class TomodachiSNSSQSEnvelopeInstance(Protocol):
    async def build_message(
        self: "TomodachiSNSSQSEnvelopeInstance", service: Any, topic: str, data: Any, **kwargs: Any
    ) -> str:
        ...

    async def parse_message(self: "TomodachiSNSSQSEnvelopeInstance", payload: str, **kwargs: Any) -> Union[dict, tuple]:
        ...


TomodachiSNSSQSEnvelope = Union[TomodachiSNSSQSEnvelopeStatic, TomodachiSNSSQSEnvelopeInstance]


async def subscribe_to(sns_client: SNSClient, sqs_client: SQSClient, topic: str, queue: str) -> None:
    """Subscribe a SQS queue to a SNS topic; create the topic and queue if they don't exist."""
    create_topic_response = await sns_client.create_topic(Name=topic)
    topic_arn = create_topic_response["TopicArn"]

    create_queue_response = await sqs_client.create_queue(QueueName=queue)
    queue_url = create_queue_response["QueueUrl"]

    get_queue_attributes_response = await sqs_client.get_queue_attributes(
        QueueUrl=queue_url, AttributeNames=["QueueArn"]
    )
    queue_arn = get_queue_attributes_response["Attributes"]["QueueArn"]

    await sns_client.subscribe(TopicArn=topic_arn, Protocol="sqs", Endpoint=queue_arn)


async def receive(
    sqs_client: SQSClient,
    queue: str,
    envelope: TomodachiSNSSQSEnvelope,
    message_type: Type[MessageType],
    max_messages: int = 10,
) -> List[MessageType]:
    """Receive messages from a SQS queue."""
    get_queue_url_response = await sqs_client.get_queue_url(QueueName=queue)
    queue_url = get_queue_url_response["QueueUrl"]

    received_messages_response = await sqs_client.receive_message(QueueUrl=queue_url, MaxNumberOfMessages=max_messages)
    received_messages = received_messages_response.get("Messages")
    if not received_messages:
        return []

    parsed_messages: List[MessageType] = []
    for received_message in received_messages:
        payload = json.loads(received_message["Body"])["Message"]
        parsed_message = await envelope.parse_message(payload=payload, proto_class=message_type)
        parsed_messages.append(parsed_message[0]["data"])
        await sqs_client.delete_message(QueueUrl=queue_url, ReceiptHandle=received_message["ReceiptHandle"])
    return parsed_messages


async def publish(sns_client: SNSClient, topic: str, data: Any, envelope: TomodachiSNSSQSEnvelopeStatic) -> None:
    """Publish a message to a SNS topic."""
    message = await envelope.build_message(service={}, topic=topic, data=data)

    list_topics_response = await sns_client.list_topics()
    topic_arn = next((v["TopicArn"] for v in list_topics_response["Topics"] if v["TopicArn"].endswith(topic)), None)
    if not topic_arn:
        raise ValueError(f"Topic does not exist: {topic}")

    await sns_client.publish(TopicArn=topic_arn, Message=message)


async def get_queue_attributes(
    sqs_client: SQSClient, queue: str, attributes: List[QueueAttributeFilterType]
) -> Dict[QueueAttributeNameType, str]:
    """Get attributes for a SQS queue."""
    get_queue_response = await sqs_client.get_queue_url(QueueName=queue)
    queue_url = get_queue_response["QueueUrl"]
    get_queue_attributes_response = await sqs_client.get_queue_attributes(QueueUrl=queue_url, AttributeNames=attributes)
    return get_queue_attributes_response["Attributes"]



"""WireMock testcontainer.

Adaptation of https://github.com/wiremock/python-wiremock
"""
from pathlib import Path
from typing import Any

from testcontainers.core.waiting_utils import wait_for_logs

from tomodachi_testcontainers.containers.common import DockerContainer, copy_folder_to_container


class WireMockContainer(DockerContainer):
    MAPPINGS_DIR: Path = Path("/home/wiremock/mappings/")
    FILES_DIR: Path = Path("/home/wiremock/__files/")

    def __init__(
        self,
        mapping_stubs: Path,
        mapping_files: Path,
        image: str = "wiremock/wiremock:latest",
        internal_port: int = 8080,
        edge_port: int = 8080,
        verbose: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(image, **kwargs)
        self.mapping_stubs = mapping_stubs
        self.mapping_files = mapping_files

        self.internal_port = internal_port
        self.edge_port = edge_port
        self.with_bind_ports(self.internal_port, self.edge_port)

        if verbose:
            self.with_command("--verbose")

    def get_internal_url(self) -> str:
        ip = self.get_container_internal_ip()
        return f"http://{ip}:{self.internal_port}"

    def get_external_url(self) -> str:
        host = self.get_container_host_ip()
        return f"http://{host}:{self.edge_port}"

    def start(self, timeout: float = 10) -> "WireMockContainer":
        super().start()
        wait_for_logs(self, "port:", timeout=timeout)
        self.copy_mappings_to_container()
        self.copy_mapping_files_to_container()
        self.reload_mappings()
        return self

    def copy_mappings_to_container(self) -> None:
        copy_folder_to_container(
            host_path=self.mapping_stubs, container_path=self.MAPPINGS_DIR, container=self.get_wrapped_container()
        )

    def copy_mapping_files_to_container(self) -> None:
        copy_folder_to_container(
            host_path=self.mapping_files, container_path=self.FILES_DIR, container=self.get_wrapped_container()
        )

    def reload_mappings(self) -> None:
        self.exec(["curl", "-X", "POST", f"http://localhost:{self.internal_port}/__admin/mappings/reset"])
